from pathlib import Path
from typing import Union

from roocs_utils.parameter import parameterise


def map_params(ds, time=None, area=None, level=None):
    """ Generates a dictionary of subset limit from parameters, which can be passed to subset """
    args = dict()

    parameters = parameterise(collection=ds, time=time, area=area, level=level)

    for parameter in ["time", "area", "level"]:

        if parameters.get(parameter).tuple is not None:
            args.update(parameters.get(parameter).asdict())

    # rename start_time and end_time to start_date and end_date to
    # match clisops/core/subset
    if "start_time" in args:
        args["start_date"] = args.pop("start_time")

    if "end_time" in args:
        args["end_date"] = args.pop("end_time")

    return args


def expand_wildcards(paths: Union[str, Path]) -> list:
    """Expand the wildcards that may be present in Paths."""
    path = Path(paths).expanduser()
    parts = path.parts[1:] if path.is_absolute() else path.parts
    return [f for f in Path(path.root).glob(str(Path("").joinpath(*parts)))]



from roocs_utils.project_utils import get_project_name
from roocs_utils.xarray_utils import xarray_utils as xu

from clisops import CONFIG
from clisops.utils.output_utils import get_format_extension


def get_file_namer(name):
    """ Returns the correct filenamer from the provided name"""
    namers = {"standard": StandardFileNamer, "simple": SimpleFileNamer}

    return namers.get(name, StandardFileNamer)


class _BaseFileNamer(object):
    """ File namer base class"""

    def __init__(self):
        self._count = 0

    def get_file_name(self, ds, fmt=None):
        """ Generate numbered file names """
        self._count += 1
        extension = get_format_extension(fmt)
        return f"output_{self._count:03d}.{extension}"


class SimpleFileNamer(_BaseFileNamer):
    """
    Simple file namer class.

    Generates numbered file names.
    """

    pass


class StandardFileNamer(SimpleFileNamer):
    """
    Standard file namer class.

    Generates file names based on input dataset.
    """

    def _get_project(self, ds):
        """ Gets the project name from the input dataset """

        try:
            return get_project_name(ds)
        except Exception:
            return None

    def get_file_name(self, ds, fmt="nc"):
        """ Constructs file name. """
        template = self._get_template(ds)

        if not template:
            # Default to parent class namer if no method found
            return super().get_file_name(ds)

        self._count += 1

        attr_defaults = CONFIG[f"project:{self._get_project(ds)}"]["attr_defaults"]
        attrs = attr_defaults.copy()
        attrs.update(ds.attrs)

        self._resolve_derived_attrs(ds, attrs, template, fmt=fmt)
        file_name = template.format(**attrs)

        return file_name

    def _get_template(self, ds):
        """ Gets template to use for output file name, based on the project of the dataset. """
        try:
            return CONFIG[f"project:{self._get_project(ds)}"]["file_name_template"]
        except KeyError:
            return None

    def _resolve_derived_attrs(self, ds, attrs, template, fmt=None):
        """
        Finds var_id, time_range and format_extension of dataset and output to
        generate output file name.
        """
        if "__derive__var_id" in template:
            attrs["__derive__var_id"] = xu.get_main_variable(ds)

        if "__derive__time_range" in template:
            attrs["__derive__time_range"] = self._get_time_range(ds)

        if "__derive__extension" in template:
            attrs["__derive__extension"] = get_format_extension(fmt)

    def _get_time_range(self, da):
        """ Finds the time range of the data in the output. """
        times = da.time.values
        return times.min().strftime("%Y%m%d") + "-" + times.max().strftime("%Y%m%d")



from pathlib import Path
from typing import List, Optional, Tuple, Union

import xarray as xr

from clisops import logging, utils
from clisops.core import subset_bbox, subset_level, subset_time
from clisops.utils.common import expand_wildcards
from clisops.utils.file_namers import get_file_namer
from clisops.utils.output_utils import get_output, get_time_slices

__all__ = [
    "subset",
]

LOGGER = logging.getLogger(__file__)


def _subset(ds, args):
    if "lon_bnds" and "lat_bnds" in args:
        # subset with space and optionally time and level
        LOGGER.debug(f"subset_bbox with parameters: {args}")
        result = subset_bbox(ds, **args)
    else:
        kwargs = {}
        valid_args = ["start_date", "end_date"]
        for arg in valid_args:
            kwargs.setdefault(arg, args.get(arg, None))

        # subset with time only
        if any(kwargs.values()):
            LOGGER.debug(f"subset_time with parameters: {kwargs}")
            result = subset_time(ds, **kwargs)

        else:
            result = ds

        kwargs = {}
        valid_args = ["first_level", "last_level"]
        for arg in valid_args:
            kwargs.setdefault(arg, args.get(arg, None))

        # subset with level only
        if any(kwargs.values()):
            LOGGER.debug(f"subset_level with parameters: {kwargs}")
            result = subset_level(ds, **kwargs)

    return result


def subset(
    ds: Union[xr.Dataset, str, Path],
    *,
    time: Optional[Union[str, Tuple[str, str]]] = None,
    area: Optional[
        Union[
            str,
            Tuple[
                Union[int, float, str],
                Union[int, float, str],
                Union[int, float, str],
                Union[int, float, str],
            ],
        ]
    ] = None,
    level: Optional[
        Union[str, Tuple[Union[int, float, str], Union[int, float, str]]]
    ] = None,
    output_dir: Optional[Union[str, Path]] = None,
    output_type="netcdf",
    split_method="time:auto",
    file_namer="standard",
) -> List[Union[xr.Dataset, str]]:
    """

    Parameters
    ----------
    ds: Union[xr.Dataset, str]
    time: Optional[Union[str, Tuple[str, str]]] = None,
    area: Optional[
        Union[
            str,
            Tuple[
                Union[int, float, str],
                Union[int, float, str],
                Union[int, float, str],
                Union[int, float, str],
            ],
        ]
    ] = None,
    level: Optional[Union[str, Tuple[Union[int, float, str], Union[int, float, str]]]] = None
    output_dir: Optional[Union[str, Path]] = None
    output_type: {"netcdf", "nc", "zarr", "xarray"}
    split_method: {"time:auto"}
    file_namer: {"standard", "simple"}

    Returns
    -------
    List[Union[xr.Dataset, str]]

    Examples
    --------
    | ds: xarray Dataset or "cmip5.output1.MOHC.HadGEM2-ES.rcp85.mon.atmos.Amon.r1i1p1.latest.tas"
    | time: ("1999-01-01T00:00:00", "2100-12-30T00:00:00") or "2085-01-01T12:00:00Z/2120-12-30T12:00:00Z"
    | area: (-5.,49.,10.,65) or "0.,49.,10.,65" or [0, 49.5, 10, 65]
    | level: (1000.,) or "1000/2000" or ("1000.50", "2000.60")
    | output_dir: "/cache/wps/procs/req0111"
    | output_type: "netcdf"
    | split_method: "time:auto"
    | file_namer: "standard"

    """
    if isinstance(ds, (str, Path)):
        ds = expand_wildcards(ds)
        if len(ds) > 1:
            ds = xr.open_mfdataset(ds, use_cftime=True, combine="by_coords")
        else:
            ds = xr.open_dataset(ds[0], use_cftime=True)

    LOGGER.debug(f"Mapping parameters: time: {time}, area: {area}, level: {level}")
    args = utils.map_params(ds, time, area, level)

    subset_ds = _subset(ds, args)

    outputs = list()
    namer = get_file_namer(file_namer)()

    time_slices = get_time_slices(subset_ds, split_method)

    for tslice in time_slices:
        result_ds = subset_ds.sel(time=slice(tslice[0], tslice[1]))
        LOGGER.info(f"Processing subset for times: {tslice}")

        output = get_output(result_ds, output_type, output_dir, namer)
        outputs.append(output)

    return outputs



import math
import os
from datetime import datetime as dt
from typing import List, Tuple, Union

import dask
import pandas as pd
import xarray as xr
from roocs_utils.utils.common import parse_size
from roocs_utils.xarray_utils import xarray_utils as xu

from clisops import CONFIG, chunk_memory_limit, logging

LOGGER = logging.getLogger(__file__)

SUPPORTED_FORMATS = {
    "netcdf": {"method": "to_netcdf", "extension": "nc"},
    "nc": {"method": "to_netcdf", "extension": "nc"},
    "zarr": {"method": "to_zarr", "extension": "zarr"},
    "xarray": {"method": None, "extension": None},
}

SUPPORTED_SPLIT_METHODS = ["time:auto"]


def check_format(fmt):
    """ Checks requested format exists. """
    if fmt not in SUPPORTED_FORMATS:
        raise KeyError(
            f'Format not recognised: "{fmt}". Must be one of: {SUPPORTED_FORMATS}.'
        )


def get_format_writer(fmt):
    """ Finds the output method for the requested output format. """
    check_format(fmt)
    return SUPPORTED_FORMATS[fmt]["method"]


def get_format_extension(fmt):
    """ Finds the extension for the requested output format. """
    check_format(fmt)
    return SUPPORTED_FORMATS[fmt]["extension"]


def _format_time(tm: Union[str, dt], fmt="%Y-%m-%d"):
    """ Convert to datetime if time is a numpy datetime """
    if not hasattr(tm, "strftime"):
        tm = pd.to_datetime(str(tm))

    return tm.strftime(fmt)


def filter_times_within(times, start=None, end=None):
    """
    Takes an array of datetimes, returning a reduced array if start or end times
    are defined and are within the main array.
    """
    filtered = []
    for tm in times:
        ft = _format_time(tm, "%Y-%m-%dT%H:%M:%S")
        if start is not None and ft < start:
            continue
        if end is not None and ft > end:
            break

        filtered.append(tm)

    return filtered


def get_da(ds):
    """ Returns xr.DataArray when format of ds may be either."""
    if isinstance(ds, xr.DataArray):
        da = ds
    else:
        var_id = xu.get_main_variable(ds)
        da = ds[var_id]

    return da


def get_time_slices(
    ds: Union[xr.Dataset, xr.DataArray],
    split_method,
    start=None,
    end=None,
    file_size_limit: str = None,
) -> List[Tuple[str, str]]:

    """

    Take an xarray Dataset or DataArray, assume it can be split on the time axis
    into a sequence of slices. Optionally, take a start and end date to specify
    a sub-slice of the main time axis.

    Use the prescribed file size limit to generate a list of
    ("YYYY-MM-DD", "YYYY-MM-DD") slices so that the output files do
    not (significantly) exceed the file size limit.

    Parameters
    ----------
    ds: Union[xr.Dataset, xr.DataArray]
    split_method
    start
    end
    file_size_limit: str
      a string specifying "<number><units>".

    Returns
    -------
    List[Tuple[str, str]]
    """

    if split_method not in SUPPORTED_SPLIT_METHODS:
        raise NotImplementedError(
            f"The split method {split_method} is not implemented."
        )

    # Use default file size limit if not provided
    if not file_size_limit:
        file_size_limit = parse_size(CONFIG["clisops:write"]["file_size_limit"])

    da = get_da(ds)

    times = filter_times_within(da.time.values, start=start, end=end)
    n_times = len(times)

    if n_times == 0:
        raise Exception(f"Zero time steps found between {start} and {end}.")

    n_slices = da.nbytes / file_size_limit
    slice_length = int(n_times // n_slices)

    if slice_length == 0:
        raise Exception("Unable to calculate slice length for splitting output files.")

    slices = []
    indx = 0
    final_indx = n_times - 1

    while indx <= final_indx:

        start_indx = indx
        indx += slice_length
        end_indx = indx - 1

        if end_indx > final_indx:
            end_indx = final_indx
        slices.append(
            (f"{_format_time(times[start_indx])}", f"{_format_time(times[end_indx])}")
        )

    return slices


def get_chunk_length(da):
    """
    Calculate the chunk length to use when chunking xarray datasets.

    Based on memory limit provided in config and the size of th dataset.
    """
    size = da.nbytes
    n_times = len(da.time.values)
    mem_limit = parse_size(chunk_memory_limit)

    if size > 0:
        n_chunks = math.ceil(size / mem_limit)
    else:
        n_chunks = 1

    chunk_length = math.ceil(n_times / n_chunks)

    return chunk_length


def _get_chunked_dataset(ds):
    """
    Chunk xr.Dataset and return chunked dataset
    """
    da = get_da(ds)
    chunk_length = get_chunk_length(da)
    chunked_ds = ds.chunk({"time": chunk_length})
    da.unify_chunks()
    return chunked_ds


def get_output(ds, output_type, output_dir, namer):

    """
    Return output after applying chunking and determining
    the output format and chunking
    """
    fmt_method = get_format_writer(output_type)
    LOGGER.info(f"fmt_method={fmt_method}, output_type={output_type}")

    if not fmt_method:
        LOGGER.info(f"Returning output as {type(ds)}")
        return ds

    file_name = namer.get_file_name(ds, fmt=output_type)
    chunked_ds = _get_chunked_dataset(ds)

    if not output_dir:
        output_dir = "."
    output_path = os.path.join(output_dir, file_name)

    # TODO: writing output works currently only in sync mode, see:
    #  - https://github.com/roocs/rook/issues/55
    #  - https://docs.dask.org/en/latest/scheduling.html
    with dask.config.set(scheduler="synchronous"):

        writer = getattr(chunked_ds, fmt_method)
        delayed_obj = writer(output_path, compute=False)
        delayed_obj.compute()

    LOGGER.info(f"Wrote output file: {output_path}")
    return output_path
